@@ -11,11 +11,6 @@
 
 /* Wishlist:
 
-   - XScreenSaver support.  The Xlib version handled the --root option and the
-     $XSCREENSAVER_WINDOW environment variable to render onto an existing X
-     window.  This is probably impossible with GTK, as there appears to be no
-     mechanism to alter the X11 Window underlying a GtkWindow.
-
    - Bifurcated transparency.  On macOS, we can render the text color with 100%
      opacity and background with less, down to 0, meaning the digits are fully
      opaque while the background is partially transparent.  It looks great.
@@ -36,6 +31,14 @@
 #include "prefs.h"
 #include "hsv.h"
 
+#include <X11/Xlib.h>
+
+#ifdef DO_SAVER
+# undef GDK_MULTIHEAD_SAFE
+# include <gdk/gdkx.h>
+#endif
+
+
 /* These aren't settable through the GUI, so no point in putting them
    in GSettings, I guess...
  */
@@ -48,6 +51,15 @@
 #include <GL/gl.h>
 #include <GL/glext.h>
 
+#ifdef DO_SAVER
+# ifdef HAVE_EGL
+#  include <EGL/egl.h>
+#  include <EGL/eglext.h>
+# else
+#  include <GL/glx.h>
+# endif
+#endif
+
 #define DO_DOUBLE_BUFFER
 
 struct _XDaliClockWindow {
@@ -73,12 +85,26 @@ struct XDaliClockWindowPrivate {
   guint date_off_timer;
   guint countdown_home_stretch_start_timer;
   guint countdown_home_stretch_end_timer;
-  double auto_date_interval;
   rgba fg, bg;
   rgba orig_fg, orig_bg;
 
   GtkGLArea *glarea;
   GLuint textures[2];
+
+# ifdef DO_SAVER
+  Display *dpy;
+  Window window_id;
+
+#  ifdef HAVE_EGL
+  EGLDisplay egl_display;
+  EGLSurface egl_surface;
+  EGLContext egl_context;
+  EGLConfig  egl_config;
+#  else
+  GLXContext glx_context;
+#  endif
+
+# endif // DO_SAVER
 };
 
 G_DEFINE_TYPE_WITH_PRIVATE (XDaliClockWindow, xdaliclock_app_window,
@@ -131,6 +157,125 @@ check_gl_error (const char *type)
 }
 
 
+
+#ifdef DO_SAVER
+
+/* Here's how XScreenSaver support *should* work: just render the GTK
+   widget hierarchy onto an external X11 window, and let it run normally:
+
+     GtkWindow *win = ...;
+     GdkDisplay *gdpy = gdk_display_get_default();
+     GdkWindow *gw = gdk_x11_window_foreign_new_for_display (gdpy, xwin);
+     gtk_widget_set_has_window (win, TRUE);
+     gdk_window_set_user_data (gw, win);
+     g_signal_connect (win, "realize", G_CALLBACK(foreign_realize), gw);
+     and in foreign_realize():
+       gtk_widget_set_window (widget, GDK_WINDOW (user_data));
+
+   Spoiler alert, that doesn't work.  I can find no evidence that creating
+   a GtkWindow and GtkWidget hierarchy atop an X11 Window works at all.
+   See the sample program in this comment on my blog for our best attempt:
+   https://www.jwz.org/blog/2022/08/dali-clock-2-46-released/#comment-236849
+
+   So instead what we do is this:
+
+     - Create a new EGL or GLX context on the existing X11 Window;
+     - Create, but do not ever realize, the GtkGLArea widget;
+     - Manually run that widget's "render" function, and have it bind
+       to our context instead of the context that the GtkGLArea would
+       have created.
+
+   It's nasty, and it's a lot more code (because EGL is insanely verbose),
+   but it works.
+ */
+static void
+init_external_glcontext (XDaliClockWindow *win)
+{
+  XDaliClockWindowPrivate *priv =
+    xdaliclock_app_window_get_instance_private (win);
+  XWindowAttributes xgwa;
+
+  if (! XGetWindowAttributes (priv->dpy, priv->window_id, &xgwa)) {
+    fprintf (stderr, "%s: bad window: 0x%lX\n", progname, priv->window_id);
+    exit (1);
+  }
+
+# ifdef HAVE_EGL
+
+  priv->egl_display = eglGetDisplay ((EGLNativeDisplayType) priv->dpy);
+  if (!priv->egl_display) {
+    fprintf (stderr, "%s: eglGetDisplay failed\n", progname);
+    exit (1);
+  }
+
+  int egl_major = -1, egl_minor = -1;
+  if (! eglInitialize (priv->egl_display, &egl_major, &egl_minor)) {
+    /* The library already printed this, but without progname:
+       "libEGL warning: DRI2: failed to create any config" */
+    /* fprintf (stderr, "%s: eglInitialize failed\n", progname); */
+    exit (1);
+  }
+
+  unsigned int vid = XVisualIDFromVisual (xgwa.visual);
+
+  const EGLint conf_attrs[] = {
+    EGL_RED_SIZE,   8,
+    EGL_GREEN_SIZE, 8,
+    EGL_BLUE_SIZE,  8,
+    EGL_RENDERABLE_TYPE, EGL_OPENGL_BIT,
+      EGL_NATIVE_VISUAL_ID, vid,
+    EGL_NONE
+  };
+  const EGLint ctx_attrs[] = {
+    EGL_CONTEXT_CLIENT_VERSION, 2,
+    EGL_NONE
+  };
+
+  int nconfig = -1;
+  if (! (eglChooseConfig (priv->egl_display, conf_attrs, &priv->egl_config,
+                          1, &nconfig) &&
+         nconfig == 1)) {
+    fprintf (stderr, "%s: no matching EGL config for X11 visual 0x%X\n",
+             progname, vid);
+    exit (1);
+  }
+
+  if (! eglBindAPI(EGL_OPENGL_API)) {
+    fprintf (stderr, "%s: eglBindAPI failed\n", progname);
+    abort();
+  }
+
+  priv->egl_context = eglCreateContext (priv->egl_display, priv->egl_config,
+                                        EGL_NO_CONTEXT, ctx_attrs);
+  if (!priv->egl_context) {
+    fprintf (stderr, "%s: eglCreateContext failed\n", progname);
+    exit (1);
+  }
+
+  priv->egl_surface =
+    eglCreatePlatformWindowSurface (priv->egl_display, priv->egl_config,
+                                    (void *) &priv->window_id, NULL);
+  if (!priv->egl_surface) {
+    fprintf (stderr, "%s: eglCreatePlatformWindowSurface failed\n",
+             progname);
+    exit (1);
+  }
+
+# else  // GLX
+
+  XVisualInfo vi_in, *vi_out;
+  int out_count;
+  vi_in.visualid = XVisualIDFromVisual (xgwa.visual);
+  vi_out = XGetVisualInfo (priv->dpy, VisualIDMask, &vi_in, &out_count);
+  if (! vi_out) abort ();
+  priv->glx_context = glXCreateContext (priv->dpy, vi_out, 0, GL_TRUE);
+  XFree ((char *) vi_out);
+
+# endif // GLX
+}
+#endif // DO_SAVER
+
+
 // size, position or stacking changed.
 //
 gboolean
@@ -144,15 +289,6 @@ gl_resize_cb (GtkGLArea *self, gint new_width, gint new_height, gpointer data)
   priv->glarea_width  = new_width;
   priv->glarea_height = new_height;
 
-# if 0
-  /* If the user is interactively resizing the window, don't regenerate
-     the bitmap until we're done.  (We will just scale whatever image is
-     already on the window instead, which reduces flicker when the
-     target bitmap size shifts over a boundary).
-   */
-  if ([self inLiveResize]) return FALSE;
-# endif
-
   int ow = config->width;
   int oh = config->height;
 
@@ -184,6 +320,61 @@ gl_resize_cb (GtkGLArea *self, gint new_width, gint new_height, gpointer data)
 }
 
 
+/* gtk_window_move() sets the origin of the window's WM decorations, but
+   GTK's "configure-event" returns the root-relative origin of the window
+   within the decorations, so the "configure-event" numbers are too large by
+   the size of the decorations (title bar and border).  Without compensating
+   for this, the window would move down and slightly to the right every time
+   we saved and restored.  GDK provides no way to find those numbers, so we
+   have to hack it out X11 style...
+
+   This probably means that under Wayland, the window will keep shifting.
+ */
+static void
+wm_decoration_origin (GtkWindow *gtkw, int *x, int *y)
+{
+  Display *dpy = gdk_x11_get_default_xdisplay();
+  GdkWindow *gdkw = gtk_widget_get_window (GTK_WIDGET (gtkw));
+  Window xw = gdk_x11_window_get_xid (gdkw);
+
+  Window root, parent, *kids;
+  unsigned int nkids;
+
+  Atom type = None;
+  int format;
+  unsigned long nitems, bytesafter;
+  unsigned char *data;
+
+  static Atom swm_vroot = 0;
+  XWindowAttributes xgwa;
+
+  if (!dpy || !xw) return;
+  if (! XQueryTree (dpy, xw, &root, &parent, &kids, &nkids))
+    abort ();
+
+  if (parent == root)	/* No window above us at all */
+    return;
+
+  if (! swm_vroot)
+    swm_vroot = XInternAtom (dpy, "__SWM_VROOT", False);
+
+  /* If parent is a virtual root, there is no intervening WM decoration. */
+  if (XGetWindowProperty (dpy, parent, swm_vroot,
+                          0, 0, False, AnyPropertyType,
+                          &type, &format, &nitems, &bytesafter,
+                          (unsigned char **) &data)
+      == Success
+      && type != None)
+    return;
+
+  /* If we have a parent, it is the WM decoration, so use its origin. */
+  if (! XGetWindowAttributes (dpy, parent, &xgwa))
+    abort();
+  *x = xgwa.x;
+  *y = xgwa.y;
+}
+
+
 // When the window is resized or moved, save the new geometry in GSettings.
 //
 gboolean
@@ -193,10 +384,13 @@ window_resize_cb (GtkWindow *window, GdkEvent *event, gpointer data)
   XDaliClockWindowPrivate *priv =
     xdaliclock_app_window_get_instance_private (win);
 
+  int x = event->configure.x;
+  int y = event->configure.y;
+  wm_decoration_origin (window, &x, &y);
+
   char geom[100];
   sprintf (geom, "%dx%d+%d+%d", 
-           event->configure.width, event->configure.height,
-           event->configure.x, event->configure.y);
+           event->configure.width, event->configure.height, x, y);
   g_settings_set_string (priv->settings, "geometry", geom);
 
   return FALSE;
@@ -211,8 +405,6 @@ gl_render_cb (GtkGLArea *area, GdkGLContext *context, gpointer data)
     xdaliclock_app_window_get_instance_private (win);
   dali_config *config = priv->config;
 
-  gboolean constrainSizes = TRUE;
-
   struct {
     struct { double width, height; } size;
     struct { double x, y; } origin;
@@ -249,14 +441,16 @@ gl_render_cb (GtkGLArea *area, GdkGLContext *context, gpointer data)
     // so limit the number size in screensaver-mode.  Note that this
     // limits the size of the output quad, not the size of the texture.
     //
-    if (constrainSizes) {
-      int maxh = (config->time_mode == SS ? 512 : /*256*/ 200);
+# ifdef DO_SAVER
+    if (priv->window_id) {
+      int maxh = (config->time_mode == SS ? 512 : 256);
       // maxh *= s;
       if (torect.size.height > maxh) {
         torect.size.height = maxh;
         torect.size.width  = maxh * img_aspect;
       }
     }
+# endif // DO_SAVER
 
     // put a 10% margin between the numbers and the edge of the window.
     torect.size.width  *= 0.9;
@@ -282,7 +476,24 @@ gl_render_cb (GtkGLArea *area, GdkGLContext *context, gpointer data)
      Set the current GL context before talking to OpenGL
    **************************************************************************/
 
-    // gtk_gl_area_make_current (area);   // Should already be current
+# ifdef DO_SAVER
+    if (priv->window_id)
+      {
+#  ifdef HAVE_EGL
+        if (! eglMakeCurrent (priv->egl_display, priv->egl_surface,
+                              priv->egl_surface, priv->egl_context))
+          abort();
+#  else  // GLX
+        if (! glXMakeCurrent (priv->dpy, priv->window_id, priv->glx_context))
+          abort();
+        glDrawBuffer (GL_BACK);
+#  endif // GLX
+      }
+    else
+# endif // DO_SAVER
+      {
+        // gtk_gl_area_make_current (area);   // Should already be current
+      }
 
   /**************************************************************************
      Bind the bitmap to a texture
@@ -386,13 +597,32 @@ gl_render_cb (GtkGLArea *area, GdkGLContext *context, gpointer data)
      Set the viewport and projection matrix
    **************************************************************************/
 
-  glMatrixMode(GL_PROJECTION);
-  glLoadIdentity();
   glViewport (0, 0, framerect.size.width, framerect.size.height);
-  glOrtho (0, framerect.size.width, 0, framerect.size.height, -1, 1);
+
+  glMatrixMode(GL_PROJECTION);
+  { // glLoadIdentity();
+    // glOrtho (0, framerect.size.width, 0, framerect.size.height, -1, 1);
+    GLfloat a = 2 / framerect.size.width;
+    GLfloat b = 2 / framerect.size.height;
+    const GLfloat m[16] = {
+      a, 0, 0, 0,
+      0, b, 0, 0,
+      0, 0,-1, 0,
+     -1,-1, 0, 1
+    };
+    glLoadMatrixf (m);
+  }
 
   glMatrixMode(GL_MODELVIEW);
-  glLoadIdentity();
+  { // glLoadIdentity();
+    const GLfloat m[16] = {
+      1, 0, 0, 0,
+      0, 1, 0, 0,
+      0, 0, 1, 0,
+      0, 0, 0, 1
+    };
+    glLoadMatrixf (m);
+  }
 
 
   /**************************************************************************
@@ -445,6 +675,16 @@ gl_render_cb (GtkGLArea *area, GdkGLContext *context, gpointer data)
      Flush bits to the screen.
    **************************************************************************/
 
+# ifdef DO_SAVER
+#  ifdef HAVE_EGL
+  if (priv->window_id)
+    eglSwapBuffers (priv->egl_display, priv->egl_surface);
+#  else   // GLX
+  if (priv->window_id)
+    glXSwapBuffers (priv->dpy, priv->window_id);
+#  endif  // GLX
+# endif // DO_SAVER
+
   glFinish();
 
   return TRUE;
@@ -627,7 +867,33 @@ clock_tick_cb (gpointer data)
     }
 
     render_once (config);
-    gtk_widget_queue_draw (GTK_WIDGET (priv->glarea));
+
+# ifdef DO_SAVER
+    if (priv->window_id)
+      {
+        static int tick = 0;
+
+        // I can't figure out how to process ConfigureNotify events,
+        // so let's just poll the window position about once a second.
+        if (++tick >= 30) {
+          tick = 0;
+          XWindowAttributes xgwa;
+          if (! XGetWindowAttributes (priv->dpy, priv->window_id, &xgwa))
+            abort();
+          if (xgwa.width  != priv->glarea_width ||
+              xgwa.height != priv->glarea_height) {
+            priv->glarea_width  = xgwa.width;
+            priv->glarea_height = xgwa.height;
+            gl_resize_cb (priv->glarea, xgwa.width, xgwa.height, win);
+          }
+        }
+
+        gl_render_cb (priv->glarea, gtk_gl_area_get_context (priv->glarea),
+                      data);
+      }
+    else
+# endif  // DO_SAVER
+      gtk_widget_queue_draw (GTK_WIDGET (priv->glarea));
   }
 
   /* Re-schedule the timer according to current fps.
@@ -762,8 +1028,6 @@ date_tick_cb (gpointer data)
   if (config->time_mode != SS)		// don't auto-date in secs-only mode
     config->display_date_p = !was_on;
 
-  /* re-schedule the timer according to current fps.
-   */
   if (!was_on) delay = 3.0;
 
   priv->date_timer = g_timeout_add (delay * 1000, date_tick_cb, win);
@@ -882,7 +1146,8 @@ gl_create_context_cb (GtkGLArea *area, gpointer data)
   // This doesn't seem to do anything.
   // "versions less than 3.2 are not supported"?
   gdk_gl_context_set_required_version (ctx, 3, 1);
-  fprintf(stderr,"## %x %x\n", gwin, ctx);
+  gtk_gl_area_set_required_version (priv->glarea, 1, 3);
+
   return ctx;
 }
 */
@@ -994,7 +1259,11 @@ xdaliclock_app_window_class_init (XDaliClockWindowClass *class)
 
 
 XDaliClockWindow *
-xdaliclock_app_window_new (XDaliClock *app, dali_config *c)
+xdaliclock_app_window_new (XDaliClock *app, dali_config *c
+# ifdef DO_SAVER
+                           , unsigned long window_id
+# endif
+                           )
 {
   XDaliClockWindow *win =
     g_object_new (XDALICLOCK_APP_WINDOW_TYPE, "application", app, NULL);
@@ -1003,21 +1272,65 @@ xdaliclock_app_window_new (XDaliClock *app, dali_config *c)
   priv->app = app;
   priv->config = c;
 
-  // Restore to the previous size/position.
-  char *geom = g_settings_get_string (priv->settings, "geometry");
-  if (geom && *geom) {
-    int w, h, x, y;
-    char c;
-    if (4 == sscanf (geom, "%d x %d + %d + %d %c", &w, &h, &x, &y, &c)) {
-      gtk_window_set_default_size (GTK_WINDOW (win), w, h);
-      gtk_window_move (GTK_WINDOW (win), x, y);
+# ifdef DO_SAVER
+  if (window_id)
+    {
+      priv->dpy = gdk_x11_get_default_xdisplay();
+
+      /* Debian 11.4, Gtk 3.24.24, 2022: under Wayland, get_default_xdisplay
+         is returning uninitialized data! */
+      if (priv->dpy &&
+          (ProtocolVersion (priv->dpy) != 11 ||
+           ProtocolRevision (priv->dpy) != 0))
+        {
+          /*
+          fprintf (stderr, "%s: uninitialized data in Display: "
+                   "protocol version %d.%d!\n", progname,
+                   ProtocolVersion(priv->dpy), ProtocolRevision(priv->dpy));
+           */
+          priv->dpy = NULL;
+        }
+
+      if (!priv->dpy) {
+        fprintf (stderr, "%s: no GTK X11 display connection (Wayland?)\n",
+                 progname);
+        exit (1);
+      }
+
+      if (window_id == ~0L)   /* caller uses -1 to mean root */
+        window_id = RootWindowOfScreen (DefaultScreenOfDisplay (priv->dpy));
+
+      priv->window_id = window_id;
+
+      XWindowAttributes xgwa;
+      if (! XGetWindowAttributes (priv->dpy, priv->window_id, &xgwa)) {
+        fprintf (stderr, "%s: bad window: 0x%lX\n", progname, priv->window_id);
+        exit (1);
+      }
+
+      // Realize the glarea but do not map it.
+      gtk_widget_realize (GTK_WIDGET (priv->glarea));
+      init_external_glcontext (win);
+      gl_resize_cb (priv->glarea, xgwa.width, xgwa.height, win);
     }
-  }
-  if (geom) free (geom);
+  else
+# endif // DO_SAVER
+    {
+      // Restore to the previous size and position.
+      char *geom = g_settings_get_string (priv->settings, "geometry");
+      if (geom && *geom) {
+        int w, h, x, y;
+        char c;
+        if (4 == sscanf (geom, "%d x %d + %d + %d %c", &w, &h, &x, &y, &c)) {
+          gtk_window_set_default_size (GTK_WINDOW (win), w, h);
+          gtk_window_move (GTK_WINDOW (win), x, y);
+        }
+      }
+      if (geom) free (geom);
 
-  gtk_window_set_decorated (GTK_WINDOW (win), !c->hide_titlebar_p);
-
-  gtk_window_set_skip_taskbar_hint (GTK_WINDOW (win), TRUE);
+      gtk_window_set_decorated (GTK_WINDOW (win), !c->hide_titlebar_p);
+      gtk_window_set_skip_taskbar_hint (GTK_WINDOW (win), TRUE);
+    }
 
   return win;
 }
